#!/usr/bin/env python3

from gi.repository import Gtk, Gio, GLib, Gdk, GdkPixbuf, Clutter, GtkClutter, GObject

import comipoli_numsort
from comipoli_pageactor import ComipoliPage
from comipoli_gobutton import ComipoliGoButton
from comipoli_thumbnail import ComipoliThumbnail
from comipoli_topbar import ComipoliTopBar
from comipoli_headerbar import ComipoliHeaderBar
from comipoli_archive import ComipoliArchive

class ComipoliWindow(Gtk.ApplicationWindow):
    def __init__(self, app):
        Gtk.ApplicationWindow.__init__(self, application=app)
        GtkClutter.init();
        # var
        self.arrow_replace = False
        self.thumbnail_height = 120
        self.num = 0
        self.archive = ComipoliArchive()
        self.is_fullscreen = False
        self.is_unrar = GLib.spawn_command_line_sync("type unrar")[3] == 0
        self.is_7za = GLib.spawn_command_line_sync("type 7za")[3] == 0
        #
        accelgroup = Gtk.AccelGroup()
        self.add_accel_group(accelgroup)
        # Icon
        iconpath = GLib.build_filenamev([GLib.path_get_dirname(__file__), "data/icons/48x48/apps/comipoli.png"])
        self.set_icon(GdkPixbuf.Pixbuf.new_from_file(iconpath))
        # Headerbar
        self.headerbar = ComipoliHeaderBar()
        self.headerbar.open_button.connect("clicked", self.on_open_clicked)
        self.headerbar.lr_button.connect("toggled", self.on_lr_button_toggled)
        self.headerbar.grid_button.connect("clicked", self.on_grid_button_clicked)
        self.set_titlebar(self.headerbar)
        # DnD
        self.drag_dest_set(
            Gtk.DestDefaults.MOTION | Gtk.DestDefaults.HIGHLIGHT | Gtk.DestDefaults.DROP,
            [Gtk.TargetEntry.new("text/uri-list", 0, 0)],
            Gdk.DragAction.MOVE
        )
        # Clutter
        embed = GtkClutter.Embed()
        self.add(embed)
        self.stage = embed.get_stage()
        self.stage.set_background_color(Clutter.Color.new(0, 0, 0, 255))
        # Page
        self.firstpage = ComipoliPage()
        self.stage.add_child(self.firstpage)
        self.secondpage = ComipoliPage()
        self.stage.add_child(self.secondpage)
        # Side Button
        self.rightbutton = ComipoliGoButton(True)
        self.rightbutton.button.connect("clicked", self.on_arrowbutton_clicked, True)
        self.stage.add_child(self.rightbutton)
        self.leftbutton = ComipoliGoButton(False)
        self.leftbutton.button.connect("clicked", self.on_arrowbutton_clicked, False)
        self.stage.add_child(self.leftbutton)
        # TopBar
        self.topbar = ComipoliTopBar()
        self.topbar.open_button.connect("clicked", self.on_open_clicked)
        self.topbar.grid_button.connect("clicked", self.on_grid_button_clicked)
        self.topbar.restore_button.connect("clicked", self.on_restore_button_clicked)
        self.topbar.lr_button.connect("toggled", self.on_lr_button_toggled)
        GObject.Binding.bind_property(
            self.headerbar.lr_button, "active",
            self.topbar.lr_button, "active",
            GObject.BindingFlags.BIDIRECTIONAL )
        self.stage.add_child(self.topbar)
        # signal
        self.stage.connect("allocation-changed", self.on_stage_allocation_changed)
        self.connect("key-press-event", self.on_key_press_event)
        self.connect("motion-notify-event", self.on_motion_notify_event)
        # settings
        try:            
            settings = Gio.Settings(schema="org.sasakima.comipoli")
            w, h = settings.get_value("window-size").unpack()
            self.resize(w, h)
            self.arrow_replace = settings.get_boolean("arrow-replace")
            self.headerbar.lr_button.props.active = settings.get_boolean("left-to-right")
            self.thumbnail_height = settings.get_int("thumbnail-height")
        except Exception as e:
            print(e)
        #
        self.show_all()

    def set_uri(self, uri):
        if uri == None:
            return
        if self.archive.new_cba(uri, self.is_unrar, self.is_7za):
            self.headerbar.set_title(self.archive.arc.get_basename())
            self.num = 0
            self.change_pixbuf(0)

    def change_pixbuf(self, next):
        if len(self.archive) == 0:
            return
        elif len(self.archive) < self.num + next + 1 and not next == 50:
            path = GLib.path_get_dirname(self.archive.path)
            extlist = ["cbz"]
            if self.is_unrar:
                extlist.append("cbr")
            if self.is_7za:
                extlist.append("cb7")
            cbzs = comipoli_numsort.get_file_numsort_list(path, extlist)
            if len(cbzs) == 0:
                return
            else:
                next_index = cbzs.index(self.archive.arc.get_basename()) + 1
                if len(cbzs) > next_index:
                    dlg = Gtk.MessageDialog(
                        transient_for=self,
                        buttons=Gtk.ButtonsType.OK_CANCEL,
                        message_type=Gtk.MessageType.QUESTION,
                        text='<span bgcolor="red">Next:</span> {0} ?'.format(cbzs[next_index]),
                        use_markup=True,
                        secondary_text="ESC Cancel")
                    dlg.set_default_response(Gtk.ResponseType.OK)
                    if dlg.run() == Gtk.ResponseType.OK:
                        uri = GLib.filename_to_uri(GLib.build_filenamev([path, cbzs[next_index]]))
                        self.set_uri(uri)
                    dlg.destroy()
        else:
            if next == 0:
                # Start Page
                self.firstpage.set_pixbuf(self.archive[0])
                self.num = 0
                self.secondpage.hide()
                self.set_pagesize()
            elif next == 50:
                # Last Page
                self.firstpage.set_pixbuf(self.archive[len(self.archive) - 1])
                self.num = len(self.archive) - 1
                self.secondpage.hide()
                self.set_pagesize()
            elif next > 0:
                # Next Page
                if self.secondpage.props.visible:
                    num = self.num + next
                else:
                    num = self.num + next - 1
                if len(self.archive) > num:
                    # Set Right Page
                    archive = self.archive[num]
                    self.firstpage.set_pixbuf(archive)
                    # Check (width > height)
                    xy1 = archive.get_width() - archive.get_height()
                    xy2 = 1
                    if len(self.archive) > num + 1:
                        archive2 = self.archive[num + 1]
                        xy2 = archive2.get_width() - archive2.get_height()
                    if xy1 > 0 or xy2 > 0:
                        # Single View
                        self.secondpage.hide()
                    else:
                        # Spread View
                        if len(self.archive) > num + 1:
                            self.secondpage.set_pixbuf(archive2)
                            self.secondpage.show()
                        else:
                            self.secondpage.hide()
                    self.num = num
                    self.set_pagesize()
            else:
                # Previos Page
                num = self.num + next
                if num < 0:
                    self.num = 0
                    self.firstpage.set_pixbuf(self.archive[self.num])
                    self.secondpage.hide()
                    self.set_pagesize()
                else:
                    # Check (width > height)
                    archive = self.archive[num]
                    xy1 = archive.get_width() - archive.get_height()
                    xy2 = 1
                    if len(self.archive) > 1:
                        archive2 = self.archive[num + 1]
                        xy2 = archive2.get_width() - archive2.get_height()
                    if xy1 > 0 or xy2 > 0:
                        # Single View
                        self.firstpage.set_pixbuf(archive2)
                        self.secondpage.hide()
                        self.num = num + 1
                    else:
                        # Spread View
                        self.firstpage.set_pixbuf(archive)
                        self.secondpage.set_pixbuf(archive2)
                        self.secondpage.show()
                        self.num = num
                    self.set_pagesize()
            # Draw Page Number
            self.headerbar.set_subtitle("{0}/{1}".format(self.num + 1, len(self.archive)))

    def set_pagesize(self):
        aw, ah = self.get_size()
        bw = self.rightbutton.props.width
        # Button
        self.rightbutton.set_size(bw, ah - bw * 2)
        self.rightbutton.set_position(aw - bw, bw)
        self.leftbutton.set_size(bw, ah - bw * 2)
        self.leftbutton.set_position(0, bw)
        # TopBar
        self.topbar.props.width = aw
        # Page
        self.firstpage.save_easing_state()
        self.secondpage.save_easing_state()
        if self.secondpage.props.visible:
            if self.headerbar.lr_button.props.active:
                leftwidth = ah * self.firstpage.aspectrate
                self.firstpage.set_size(leftwidth , ah)
                self.firstpage.set_position(aw / 2 - leftwidth, 0)
                rightwidth = ah * self.secondpage.aspectrate
                self.secondpage.set_size(rightwidth , ah)
                self.secondpage.set_position(aw / 2, 0)       
            else:
                leftwidth = ah * self.secondpage.aspectrate
                self.secondpage.set_size(leftwidth , ah)
                self.secondpage.set_position(aw / 2 - leftwidth, 0)
                rightwidth = ah * self.firstpage.aspectrate
                self.firstpage.set_size(rightwidth , ah)
                self.firstpage.set_position(aw / 2, 0)
        else:
            if len(self.archive) > 0:
                w = self.archive[self.num].get_width()
                h = self.archive[self.num].get_height()
                if aw * h > ah * w:
                    width = w * ah / h
                    height = ah
                    x = (aw - width) / 2
                    y = 0
                else:
                    width = aw
                    height = h * aw / w
                    x = 0
                    y = (ah - height) / 2
                self.firstpage.set_size(width , height)
                self.firstpage.set_position(x, y)
        self.firstpage.restore_easing_state()
        self.secondpage.restore_easing_state()

    def on_lr_button_toggled(self, button):
        self.set_pagesize()
        if button.props.active:
            button.set_label("L->R")
        else:
            button.set_label("L<-R")

    def on_stage_allocation_changed(self, actor, box, flags):
        if len(self.archive) > 0:
            self.set_pagesize()

    def on_motion_notify_event(self, widget, event):
        if event.y < self.topbar.props.height:
            if self.topbar.props.opacity == 0x00:
                if self.is_fullscreen:
                    self.topbar.props.opacity = 0xaa
        else:
            if self.topbar.props.opacity == 0xaa:
                self.topbar.props.opacity = 0x00

    def on_grid_button_clicked(self, button):
        dlg = ComipoliThumbnail(self)
        dlg.run()
        dlg.destroy()

    def on_open_clicked(self, button):
        dlg = Gtk.FileChooserNative(
            title="Open",
            transient_for=self,
            action=Gtk.FileChooserAction.OPEN)
        ft = Gtk.FileFilter()
        ft.set_name("Comic Book Archive")
        ft.add_mime_type("application/x-cbz")
        ft.add_mime_type("application/vnd.comicbook+zip")
        if self.is_unrar:
            ft.add_mime_type("application/x-cbr")
        if self.is_7za:
            ft.add_mime_type("application/x-cb7")
        dlg.add_filter(ft)
        if dlg.run() == Gtk.ResponseType.ACCEPT:
            self.set_uri(dlg.get_uri())
        dlg.destroy()

    def on_arrowbutton_clicked(self, button, right):
        kmp = Gdk.Keymap.get_default()
        event = Gdk.EventKey()
        event.state = kmp.get_modifier_state()
        if right:
            event.keyval = Gdk.KEY_Right
        else:
            event.keyval = Gdk.KEY_Left
        self.on_key_press_event(None, event)

    def on_key_press_event(self, widget, event):
        """
            EoG Like Key Bind
        """
        if event.state == Gdk.ModifierType.CONTROL_MASK:
            # Step Page
            if event.keyval == Gdk.KEY_Down:
                self.change_pixbuf(1)
            elif event.keyval == Gdk.KEY_space:
                self.change_pixbuf(1)
            elif event.keyval == Gdk.KEY_Up:
                self.change_pixbuf(-1)
            elif event.keyval == Gdk.KEY_BackSpace:
                self.change_pixbuf(-1)
            # Preference
            elif event.keyval == Gdk.KEY_Right:
                if self.arrow_replace:
                    self.change_pixbuf(-1)   
                else:
                    self.change_pixbuf(1)
            elif event.keyval == Gdk.KEY_Left:
                if self.arrow_replace:
                    self.change_pixbuf(1)   
                else:
                    self.change_pixbuf(-1)
            # Open
            elif event.keyval == Gdk.KEY_o:
                self.on_open_clicked(None)
        else:
            # Next Page
            if event.keyval == Gdk.KEY_Down:
                self.change_pixbuf(2)
            elif event.keyval == Gdk.KEY_space:
                self.change_pixbuf(2)

            elif event.keyval == Gdk.KEY_Up:
                self.change_pixbuf(-2)
            elif event.keyval == Gdk.KEY_BackSpace:
                self.change_pixbuf(-2)
            # Preference
            elif event.keyval == Gdk.KEY_Right:
                if self.arrow_replace:
                    self.change_pixbuf(-2)   
                else:
                    self.change_pixbuf(2)
            elif event.keyval == Gdk.KEY_Left:
                if self.arrow_replace:
                    self.change_pixbuf(2)   
                else:
                    self.change_pixbuf(-2)
            # Home, End
            elif event.keyval == Gdk.KEY_Home:
                self.change_pixbuf(0)
            elif event.keyval == Gdk.KEY_End:
                self.change_pixbuf(50)
            # etc
            elif event.keyval == Gdk.KEY_F11:
                if self.is_fullscreen:
                    self.unfullscreen()
                    self.is_fullscreen = False
                else:
                    self.fullscreen()
                    self.is_fullscreen = True
            elif event.keyval == Gdk.KEY_Escape:
                if self.is_fullscreen:
                    self.unfullscreen()
                    self.is_fullscreen = False
                else:
                    self.close()
            elif event.keyval == Gdk.KEY_F9:
                self.on_grid_button_clicked(None)
            elif event.keyval == Gdk.KEY_F8:
                self.headerbar.lr_button.props.active = not self.headerbar.lr_button.props.active
        return False

    def on_restore_button_clicked(self, widget):
        self.unfullscreen()
        self.is_fullscreen = False
        self.topbar.props.opacity = 0x00

    def do_button_press_event(self, event):
        if event.type == Gdk.EventType.DOUBLE_BUTTON_PRESS:
            if 32 < event.x and event.x < self.get_size()[0] - 32:
                if self.is_fullscreen:
                    self.unfullscreen()
                    self.is_fullscreen = False
                else:
                    self.fullscreen()
                    self.is_fullscreen = True

    def do_drag_data_received(self, drag_context, x, y, data, info, time):
        drop = data.get_uris()[0]
        self.set_uri(drop)
        self.present()

    def do_hide(self):
        try:
            settings = Gio.Settings(schema="org.sasakima.comipoli")
            xy = GLib.Variant("(ii)", self.get_size())
            settings.set_value("window-size", xy)
            settings.set_boolean("arrow-replace", self.arrow_replace)
            settings.set_int("thumbnail-height", self.thumbnail_height)
            settings.set_boolean("left-to-right", self.headerbar.lr_button.props.active)
        except Exception as e:
            print(e)
